# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from gaebusiness.business import Command
from gaebusiness.gaeutil import ModelSearchCommand
from gaecookie import facade as cookie_facade
from gaegraph.business_base import NodeSearch
from gaepermission import inspector
from gaepermission.base_commands import UpdateUserGroups, GetMainUserByEmail, SaveUserCmd
from gaepermission.base_commands2 import LoginCheckingEmail
from gaepermission.facebook.commands import GetFacebookApp, SaveOrUpdateFacebookApp, LogFacebookUserIn, FetchFacebook
from gaepermission.google.commands import GoogleLogin
from gaepermission.model import MainUser
from gaepermission.passwordless.commands import SaveOrUpdateApp, GetApp, SengLoginEmail, Login
from tekton import router

USER_COOKIE_NAME = 'userck'


def save_user_cmd(email, name=None, groups=None, locale='en_US', timezone='US/Eastern'):
    """
    Command to save a user
    :param email: user email
    :param name: user name
    :param groups: user permission groups
    :return: A command that validate date and save the user
    """
    if name is None:
        name = email
    if groups is None:
        groups = []
    return SaveUserCmd(name=name, email=email, groups=groups, locale=locale, timezone=timezone)


def get_user_by_email(email):
    """
    Returns a command that find MainUser by her email address
    :param email: email to use in search
    :return: Command that look for user on DB
    """
    return GetMainUserByEmail(email)


def web_path_security_info():
    """
    Returns a generator that returns all paths from the application if information about groups and csrf security
    """
    return inspector.web_paths_security_info(router.package_base)


def logout(response):
    """
    Returns a command that log the user out, removing her id from cookie
    """
    return cookie_facade.delete_cookie(response, USER_COOKIE_NAME)


def logged_user(request):
    """
    Returns a command that retrieves the current logged user based on secure cookie
    If there is no logged user, the result from command is None
    """
    dct = cookie_facade.retrive_cookie_data(request, USER_COOKIE_NAME).execute().result
    if dct is None:
        return Command()
    return NodeSearch(dct['id'])


def login_google(google_user, response):
    """
    Google user must be the user returned from get_current_user from users module provided by App Engine
    Returns a command that log user in based on her google account credentials.
    The logged user (MainUser) is provides on result or None if the user is not logged with her Google Account
    """

    return GoogleLogin(google_user, response, USER_COOKIE_NAME)


def login_passwordless(ticket, response, detail_url='https://pswdless.appspot.com/rest/detail'):
    """
    Log user in using Passwordless service
    :param ticket: ticket returned from Passwordless
    :param response: Response object from webapp2
    :param detail_url: url to check ticket and user data
    :return: a Command that log user in when executed
    """
    return Login(ticket, response, USER_COOKIE_NAME, detail_url)


def login_checking_email(pending_id, ticket, response, detail_url='https://pswdless.appspot.com/rest/detail'):
    """
    Log user in using Passwordless service
    :param pending_id: PendingExternalToMainUser's id
    :param ticket: ticket returned from Passwordless
    :param response: Response object from webapp2
    :param detail_url: url to check ticket and user data
    :return: a Command that log user in when executed
    """
    return LoginCheckingEmail(pending_id, ticket, response, USER_COOKIE_NAME, detail_url)


def update_user_groups(user_id, groups):
    """
    Returns a command that updates user's groups of respective user_id.
    """
    return UpdateUserGroups(user_id, groups)


def find_users_by_email_starting_with(email_prefix=None, cursor=None, page_size=30):
    """
    Returns a command that retrieves users by its email_prefix, ordered by email.
    It returns a max number of users defined by page_size arg. Next result can be retrieved using cursor, in
    a next call. It is provided in cursor attribute from command.
    """
    email_prefix = email_prefix or ''

    return ModelSearchCommand(MainUser.query_email_starts_with(email_prefix),
                              page_size, cursor, cache_begin=None)


def find_users_by_email_and_group(email_prefix=None, group=None, cursor=None, page_size=30):
    """
    Returns a command that retrieves users by its email_prefix, ordered by email and by Group.
    If Group is None, only users without any group are going to be searched
    It returns a max number of users defined by page_size arg. Next result can be retrieved using cursor, in
    a next call. It is provided in cursor attribute from command.
    """
    email_prefix = email_prefix or ''

    return ModelSearchCommand(MainUser.query_email_and_group(email_prefix, group),
                              page_size, cursor, cache_begin=None)



def send_passwordless_login_link(email, return_url, lang='en_US', url_login='https://pswdless.appspot.com/rest/login'):
    """

    :param app_id: The Passwordless app's id
    :param token: The Passwordless app's token
    :param return_url: The url user will be redirected after clicking login link
    :return: command that communicate with passsworless to sent the email
    """
    return SengLoginEmail(email, return_url, lang, url_login)


def save_or_update_passwordless_app_data(id=None, token=None):
    """
    :param id: The App's id
    :param token: The App's token
    :return: a command that save or update existing Passwordless App Data
    See https://pswdless.appspot.com/api#register-sites
    """
    return SaveOrUpdateApp(id, token)


def get_passwordless_app_data():
    """
    :return: a command that returns the Passwordless App Data from db
    """
    return GetApp()


def get_facebook_app_data():
    """
    :return: a command that returns the Facebook App Data from db
    """
    return GetFacebookApp()


def save_or_update_facebook_app_data(id=None, token=None):
    """
    :param id: The App's id
    :param token: The App's token
    :return: a command that save or update existing Facebook App Data
    See https://developers.facebook.com/docs/facebook-login/manually-build-a-login-flow/v2.0
    """
    return SaveOrUpdateFacebookApp(id, token)



def login_facebook(token, response):
    """

    :param token: facebook request token
    :param response: http response from webapp2
    :return: a command that log the facebook user in
    """
    return LogFacebookUserIn(token, response, USER_COOKIE_NAME)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import urllib
from gaepermission import facade
from gaepermission.decorator import has_permission
from tekton.gae.middleware import Middleware

LOGIN_PATH = '/login'
LOGOUT_PATH = '/logout'


class LoggedUserMiddleware(Middleware):
    def set_up(self):
        user = facade.logged_user(self.handler.request).execute().result
        self.dependencies['_logged_user'] = user
        if user:
            self.dependencies['_logout_path'] = LOGOUT_PATH
            self.dependencies['_login_path'] = None
        else:
            return_path = urllib.urlencode({'ret_path': self.handler.request.path_qs})
            self.dependencies['_login_path'] = "%s?%s" % (LOGIN_PATH, return_path)
            self.dependencies['_logout_path'] = None


class PermissionMiddleware(Middleware):
    def set_up(self):
        fcn = self.dependencies['_fcn']
        user = self.dependencies['_logged_user']
        if not has_permission(user, fcn):
            if user is None:
                self.handler.redirect(self.dependencies['_login_path'])
            else:
                self.handler.response.status_int = 403
                self.handler.response.write('You have no access permission')
            return True


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from google.appengine.api import users
from tekton import router

_security_permission_group_map = {}


class _MsgHelper(object):
    '''
    Class that aways returns True for in operation
    '''

    def __init__(self, msg):
        self.msg = msg

    def __unicode__(self):
        return self.msg


_login_required = _MsgHelper('Login Required')
_login_not_required_helper = _MsgHelper('Login not Required')


def permissions(*groups):
    def decorator(fcn):
        path = router.to_path(fcn)
        _security_permission_group_map[path] = frozenset(groups)
        return fcn

    return decorator


def login_not_required(fcn):
    path = router.to_path(fcn)
    _security_permission_group_map[path] = _login_not_required_helper
    return fcn


def login_required(fcn):
    path = router.to_path(fcn)
    _security_permission_group_map[path] = _login_required
    return fcn


def get_groups_by_path(path):
    return _security_permission_group_map.get(path)


def get_groups(fcn):
    path = router.to_path(fcn)
    return get_groups_by_path(path)


def has_permission(user, fcn):
    groups = get_groups(fcn)

    if groups is _login_not_required_helper:
        return True
    if user and groups is _login_required:
        return True

    if user and groups:
        for g in user.groups:
            if g in groups:
                return True

    return users.is_current_user_admin()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from google.appengine.ext import ndb
from gaegraph.model import Node


class PasswordlessApp(Node):
    '''
     Class that hold data from Passwordless
     See https://pswdless.appspot.com/api#register-sites
    '''
    app_id = ndb.StringProperty(required=True, indexed=False)
    token = ndb.StringProperty(required=True, indexed=False)

# class PasswordlessLoginToken(Node):
#     '''
#     Class to hold login sent token, so it is possible certify login calls
#     '''
#     token = ndb.StringProperty(required=True, indexed=False)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json

from google.appengine.api import memcache
from google.appengine.ext import ndb
from gaebusiness.business import CommandParallel

from gaebusiness.gaeutil import ModelSearchCommand, UrlFetchCommand
from gaecookie import facade
from gaegraph.business_base import SingleDestinationSearch
from gaepermission.base_commands import GetMainUserByEmail
from gaepermission.model import MainUser, ExternalToMainUser, PasswordlessUser
from gaepermission.passwordless.model import PasswordlessApp


class GetApp(ModelSearchCommand):
    def __init__(self):
        super(GetApp, self).__init__(PasswordlessApp.query(), 1)

    def do_business(self):
        super(GetApp, self).do_business()
        if self.result:
            self.result = self.result[0]
        else:
            self.result = None


class SaveOrUpdateApp(GetApp):
    def __init__(self, id=None, token=None):
        super(SaveOrUpdateApp, self).__init__()
        self.token = token
        self.id = id

    def do_business(self):
        super(SaveOrUpdateApp, self).do_business()
        if not self.result:
            self.result = PasswordlessApp()
        else:
            memcache.delete(self._cache_key())
        if self.id:
            self.result.app_id = self.id
        if self.token:
            self.result.token = self.token

    def commit(self):
        return self.result


class SengLoginEmail(GetApp):
    def __init__(self, email, return_url, lang, url_passwordless_login):
        super(SengLoginEmail, self).__init__()
        self.url_passwordless_login = url_passwordless_login
        self.email = email
        self.return_url = return_url
        self.lang = lang


    def do_business(self):
        super(SengLoginEmail, self).do_business()
        app = self.result
        if app:
            fetch_command = UrlFetchCommand(self.url_passwordless_login,
                                            {'email': self.email,
                                             'app_id': app.app_id,
                                             'token': app.token,
                                             'hook': self.return_url,
                                             'lang': self.lang},
                                            method='POST')
            fetch_command.execute()
        else:
            self.add_error('app_data', 'Must save Passwordless App Credentials before login calls')


class GetPasswordlessUser(ModelSearchCommand):
    def __init__(self, passwordless_id):
        super(GetPasswordlessUser, self).__init__(PasswordlessUser.query_by_external_id(passwordless_id), 1)

    def do_business(self):
        super(GetPasswordlessUser, self).do_business()
        self.result = self.result[0] if self.result else None


class PasswordlessDetailCheck(GetApp):
    def __init__(self, ticket, detail_url):
        super(PasswordlessDetailCheck, self).__init__()
        self.detail_url = detail_url
        self.ticket = ticket
        self._to_commit = None

    def do_business(self):
        super(PasswordlessDetailCheck, self).do_business()
        if not self.errors:
            app = self.result
            self.result = None
            fetch_cmd = UrlFetchCommand(self.detail_url,
                                        {'ticket': self.ticket, 'app_id': app.app_id, 'token': app.token},
                                        'POST').execute()
            self.errors.update(fetch_cmd.errors)
            if not self.errors:
                dct = json.loads(fetch_cmd.result.content)
                cmd_parallel = CommandParallel(GetMainUserByEmail(dct['email']) , GetPasswordlessUser(dct['id']))
                cmd_parallel.execute()
                main_user, passwordless_user = cmd_parallel[0].result, cmd_parallel[1].result
                if passwordless_user:
                    main_user = SingleDestinationSearch(ExternalToMainUser, passwordless_user).execute().result
                elif main_user:
                    passwordless_user_key = PasswordlessUser(external_id=dct['id']).put()
                    self._to_commit = ExternalToMainUser(origin=passwordless_user_key, destination=main_user.key)
                else:
                    main_user = MainUser(name=dct['email'], email=dct['email'])
                    passwordless_user = PasswordlessUser(external_id=dct['id'])
                    ndb.put_multi([main_user, passwordless_user])
                    self._to_commit = ExternalToMainUser(origin=passwordless_user.key, destination=main_user.key)
                self.result = main_user

    def commit(self):
        return self._to_commit


class Login(PasswordlessDetailCheck):
    def __init__(self, ticket, response, user_cookie_name, detail_url):
        super(Login, self).__init__(ticket, detail_url)
        self.response = response
        self.user_cookie_name = user_cookie_name

    def do_business(self):
        super(Login, self).do_business()
        facade.write_cookie(self.response, self.user_cookie_name, {'id': self.result.key.id()}).execute()






# -*- coding: utf-8 -*-


from __future__ import absolute_import, unicode_literals


# This class should be on base_commands, but it would cause circular dependency
from gaebusiness.business import CommandParallel
from gaegraph.business_base import NodeSearch, SingleDestinationSearch
from gaepermission.model import ExternalToMainUser
from gaepermission.passwordless.commands import  Login


def _is_there_a_link_already(pending_model):
    return SingleDestinationSearch(ExternalToMainUser, pending_model.external_user).execute().result


def _is_pending_user_same_as_loging_in(main_user, pending_model):
    return pending_model.main_user == main_user.key


def _should_create_link(main_user, pending_model):
    return pending_model and _is_pending_user_same_as_loging_in(main_user,
                                                                pending_model) and not _is_there_a_link_already(
        pending_model)


class LoginCheckingEmail(CommandParallel):
    def __init__(self, pending_id, ticket, response, user_cookie_name, detail_url):
        super(LoginCheckingEmail, self).__init__(NodeSearch(pending_id),
                                                 Login(ticket, response, user_cookie_name, detail_url))
        self.checked = False
        self.__to_commit = None


    def do_business(self):
        super(LoginCheckingEmail, self).do_business()
        pending_model = self[0].result
        main_user = self[1].result
        if _should_create_link(main_user, pending_model):
            self.checked = True
            self.__to_commit = ExternalToMainUser(origin=pending_model.external_user,
                                                  destination=pending_model.main_user)

    def commit(self):
        models = super(LoginCheckingEmail, self).commit()
        if self.__to_commit:
            models.append(self.__to_commit)
        return models







# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

from gaepermission.base_commands import Login
from gaepermission.model import GoogleUser


class GoogleLogin(Login):
    def __init__(self, google_api_user, response, user_cookie):
        super(GoogleLogin, self).__init__(GoogleUser,
                                          google_api_user.user_id(),
                                          google_api_user.email(),
                                          google_api_user.nickname(),
                                          response,
                                          user_cookie)

        []


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import importlib
import inspect
from gaecookie.decorator import is_csrf_secure_by_path
from gaepermission.decorator import get_groups_by_path
import os
from tekton import router


def _paths_helper(base_path, prefix_path):
    prefix_len = len(prefix_path) + 1
    for dirpath, dirnames, filenames in os.walk(base_path, False):
        package = dirpath[prefix_len:].replace(r'/', '.').replace(os.path.sep, '.')  # second replace due windows
        pyfile_names = (f[:-3] for f in filenames if f.endswith('.py') and f != '__init__.py')
        for fname in pyfile_names:
            module_name = '.'.join((package, fname))
            module = importlib.import_module(module_name)
            for item_name in dir(module):
                item = getattr(module, item_name)
                if inspect.isfunction(item):
                    path = router.to_path(item)
                    if path.startswith('/'):
                        yield path


def web_paths(base_package):
    module = importlib.import_module(base_package)
    base_path = os.path.dirname(module.__file__)
    prefix_path = os.path.join(base_path, '..' * (base_package.count(r'.') + 1))
    prefix_path = os.path.abspath(prefix_path)
    return _paths_helper(base_path, prefix_path)


class PathInfo(object):
    def __init__(self, path):
        self.csrf = is_csrf_secure_by_path(path)
        g = get_groups_by_path(path)
        if g is None:
            g = 'SYS_OWNER'
        if isinstance(g, frozenset):
            g = ', '.join(g)
        self.groups = unicode(g)
        self.path = path


def web_paths_security_info(base_package):
    for path in web_paths(base_package):
        yield PathInfo(path)







# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from google.appengine.ext import ndb
from gaegraph.model import Node


class FacebookApp(Node):
    """Class that hold data from Facebook App"""
    app_id = ndb.StringProperty(required=True, indexed=False)
    token = ndb.StringProperty(required=True, indexed=False)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from google.appengine.ext import ndb
from gaeforms.ndb.property import Email
from gaegraph.model import Node, Arc
from google.appengine.ext.ndb.model import ComputedProperty

_USER_WITHOUT_GROUP = 'USER_WITHOUT_GROUP'


class MainUser(Node):
    name = ndb.StringProperty(required=True)
    email = Email(required=True)
    groups = ndb.StringProperty(repeated=True)
    locale = ndb.StringProperty(indexed=False)
    timezone = ndb.StringProperty(indexed=False)
    group_flag = ComputedProperty(lambda user: None if user.groups else _USER_WITHOUT_GROUP)

    @classmethod
    def _calculate_prefix(cls, prefix):
        last_str_with_prefix = prefix + unichr(65525)  # this is the last unichar supported on windows systems
        return last_str_with_prefix

    @classmethod
    def query_email_starts_with(cls, prefix=''):
        last_str_with_prefix = cls._calculate_prefix(prefix)
        return cls.query(cls.email >= prefix, cls.email < last_str_with_prefix).order(cls.email)


    @classmethod
    def query_email_and_group(cls, prefix, group):
        last_str_with_prefix = cls._calculate_prefix(prefix)
        if group is None:
            return cls.query(cls.email >= prefix, cls.email < last_str_with_prefix,
                             cls.group_flag == _USER_WITHOUT_GROUP).order(cls.email)
        return cls.query(cls.email >= prefix, cls.email < last_str_with_prefix, cls.groups == group).order(cls.email)


    @classmethod
    def query_email(cls, email):
        return cls.query(cls.email == email)


# Users from external providers

class ExternalUser(Node):
    external_id = ndb.StringProperty(required=True)

    @classmethod
    def query_by_external_id(cls, external_id):
        return cls.query(cls.external_id == external_id)


class ExternalToMainUser(Arc):
    destination = ndb.KeyProperty(MainUser, required=True)
    origin = ndb.KeyProperty(ExternalUser, required=True)


class PendingExternalToMainUser(Node):
    """
    Class used to create ExternalToMainUser after email confirmation
    """
    main_user = ndb.KeyProperty(MainUser, required=True, indexed=False)
    external_user = ndb.KeyProperty(required=True, indexed=False)
    # name = ndb.StringProperty(required=True, indexed=False)
    # email = ndb.StringProperty(required=True, indexed=False)


class GoogleUser(ExternalUser):
    pass


class PasswordlessUser(ExternalUser):
    pass


class FacebookUser(ExternalUser):
    pass







# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import json

from google.appengine.api import memcache
from google.appengine.ext import ndb
from gaebusiness.business import Command

from gaebusiness.gaeutil import ModelSearchCommand, UrlFetchCommand
from gaecookie import facade
from gaepermission.base_commands import FindMainUserFromExternalUserId, CheckMainUserEmailConflict, Login
from gaepermission.facebook.model import FacebookApp
from gaepermission.model import FacebookUser, ExternalToMainUser, MainUser


class GetFacebookApp(ModelSearchCommand):
    def __init__(self):
        super(GetFacebookApp, self).__init__(FacebookApp.query(), 1)

    def do_business(self):
        super(GetFacebookApp, self).do_business()
        if self.result:
            self.result = self.result[0]
        else:
            self.result = None


class SaveOrUpdateFacebookApp(GetFacebookApp):
    def __init__(self, id=None, token=None):
        super(SaveOrUpdateFacebookApp, self).__init__()
        self.token = token
        self.id = id

    def do_business(self):
        super(SaveOrUpdateFacebookApp, self).do_business()
        if not self.result:
            self.result = FacebookApp()
        else:
            memcache.delete(self._cache_key())
        if self.id:
            self.result.app_id = self.id
        if self.token:
            self.result.token = self.token

    def commit(self):
        return self.result


class FetchFacebook(UrlFetchCommand):
    def __init__(self, token):
        super(FetchFacebook, self).__init__('https://graph.facebook.com/v2.0/me',
                                            {'fields': 'id,email,name',
                                             'access_token': token})
        self._to_commit = None

    def do_business(self):
        super(FetchFacebook, self).do_business()
        if not self.errors:
            self.result = json.loads(self.result.content)


class LogFacebookUserIn(Command):
    def __init__(self, token, response, user_cookie_name):
        super(LogFacebookUserIn, self).__init__()
        self._fetch_facebook = FetchFacebook(token)
        self._login_cmd = None
        self.user_cookie_name = user_cookie_name
        self.response = response
        self.pending_link = None
        self.external_user = None
        self.main_user_from_external = None
        self.main_user_from_email = None

    def set_up(self):
        self._fetch_facebook.set_up()


    def do_business(self):
        self._fetch_facebook.do_business()
        self.errors.update(self._fetch_facebook.errors)
        if not self.errors:
            dct = self._fetch_facebook.result
            self._login_cmd = Login(FacebookUser,
                                    dct['id'],
                                    dct['email'],
                                    dct['name'],
                                    self.response,
                                    self.user_cookie_name)
            self._login_cmd.set_up()
            self._login_cmd.do_business()

    def commit(self):
        if self._login_cmd:
            self.result = self._login_cmd.result
            self.pending_link = self._login_cmd.pending_link
            self.external_user = self._login_cmd.external_user
            self.main_user_from_external = self._login_cmd.main_user_from_external
            self.main_user_from_email = self._login_cmd.main_user_from_email
            return self._login_cmd.commit()




# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
from google.appengine.ext import ndb
from gaebusiness.business import Command, CommandParallel
from gaebusiness.gaeutil import ModelSearchCommand, SaveCommand
from gaecookie import facade
from gaeforms.ndb.form import ModelForm
from gaegraph.business_base import NodeSearch, SingleDestinationSearch
from gaepermission.model import MainUser, ExternalToMainUser, PendingExternalToMainUser



def log_main_user_in(main_user, response, user_cookie):
    facade.write_cookie(response, user_cookie, {'id': main_user.key.id()}).execute()


class GetMainUserByEmail(ModelSearchCommand):
    def __init__(self, email):
        super(GetMainUserByEmail, self).__init__(MainUser.query_email(email), 1)

    def do_business(self):
        super(GetMainUserByEmail, self).do_business()
        self.result = self.result[0] if self.result else None


class MainUserForm(ModelForm):
    _model_class = MainUser
    _include = [MainUser.email, MainUser.name, MainUser.groups, MainUser.locale, MainUser.timezone]


class SaveUserCmd(SaveCommand):
    _model_form_class = MainUserForm

    def handle_previous(self, command):
        self.result = command.result

    def __init__(self, **form_parameters):
        super(SaveUserCmd, self).__init__(**form_parameters)

    def do_business(self):
        if self.result is None:
            super(SaveUserCmd, self).do_business()


class UpdateUserGroups(NodeSearch):
    def __init__(self, user_id, groups):
        super(UpdateUserGroups, self).__init__(user_id)
        self.groups = groups

    def do_business(self):
        super(UpdateUserGroups, self).do_business()
        self.result.groups = self.groups

    def commit(self):
        return self.result


class FindMainUserFromExternalUserId(ModelSearchCommand):
    def __init__(self, external_user_class, external_id):
        super(FindMainUserFromExternalUserId,
              self).__init__(external_user_class.query_by_external_id(external_id), 1)
        self.external_user = None

    def do_business(self):
        super(FindMainUserFromExternalUserId, self).do_business()
        external_user = self.result[0] if self.result else None
        if external_user:
            self.result = SingleDestinationSearch(ExternalToMainUser, external_user.key).execute().result
            self.external_user = external_user
        else:
            self.result = None


class CheckMainUserEmailConflict(CommandParallel):
    def __init__(self, external_user_class, external_id, email):
        super(CheckMainUserEmailConflict, self).__init__(GetMainUserByEmail(email),
                                                         FindMainUserFromExternalUserId(external_user_class,
                                                                                        external_id))
        self.external_user = None
        self.main_user_from_external = None
        self.main_user_from_email = None


    def do_business(self):
        super(CheckMainUserEmailConflict, self).do_business()
        self.result = True
        self.external_user = self[1].external_user
        self.main_user_from_external = self[1].result
        self.main_user_from_email = self[0].result

        if self.main_user_from_email and not self.main_user_from_external:
            self.result = False


class Login(CheckMainUserEmailConflict):
    def __init__(self, external_user_class, external_id, email, user_name, response, user_cookie):
        super(Login, self).__init__(external_user_class, external_id, email)
        self.user_name = user_name
        self.user_cookie = user_cookie
        self.response = response
        self._arc = None
        self.pending_link = None
        self.external_user_class = external_user_class
        self.external_id = external_id
        self.email = email


    def do_business(self):
        super(Login, self).do_business()

        # if no conflict
        if self.result:
            if self.external_user is None and self.main_user_from_external is None:
                self.external_user = self.external_user_class(external_id=self.external_id)
                self.main_user_from_external = MainUser(name=self.user_name,
                                                        email=self.email)
                external_user_key, main_user_key = ndb.put_multi([self.external_user, self.main_user_from_external])
                self._arc = ExternalToMainUser(origin=external_user_key, destination=main_user_key)
            log_main_user_in(self.main_user_from_external, self.response, self.user_cookie)
        else:
            if self.external_user is None:
                self.external_user = self.external_user_class(external_id=self.external_id)
                self.external_user.put()
            self.pending_link = PendingExternalToMainUser(external_user=self.external_user.key,
                                                          main_user=self.main_user_from_email.key)

    def commit(self):
        return [m for m in [self._arc, self.pending_link] if m]
